package main

import (
	"bytes"
	"context"
	"html/template"
	"log"
	"net/http"

	"github.com/gorilla/websocket"
	"github.com/grpc-ecosystem/grpc-gateway/v2/runtime"
	"google.golang.org/grpc"
	protojson "google.golang.org/protobuf/encoding/protojson"

	gw "git.alfi.li/gamelang/proto_gorm/proto/gen/go"
)

var upgrader = websocket.Upgrader{} // use default options

type WSMsg struct {
	Kind    string
	Payload []byte
}

type ResponseWriter struct {
	header http.Header
	data   *[]byte
}

func NewResponseWriter() *ResponseWriter {
	nrw := ResponseWriter{header: http.Header{}, data: &[]byte{}}
	return &nrw
}

func (rw ResponseWriter) Header() http.Header {
	return rw.header
}

func (rw ResponseWriter) Write(input []byte) (int, error) {
	*rw.data = input
	return http.StatusOK, nil
}

func (rw ResponseWriter) WriteHeader(statusCode int) {}

func (rw ResponseWriter) GetData() []byte {
	return *rw.data
}

type WSGateway struct {
	mux *runtime.ServeMux
}

func (wsg *WSGateway) Handler(w http.ResponseWriter, r *http.Request, pathParams map[string]string) {
	c, err := upgrader.Upgrade(w, r, nil)
	if err != nil {
		log.Print("upgrade:", err)
		return
	}
	defer c.Close()
	for {
		mt, message, err := c.ReadMessage()
		if err != nil {
			log.Println("read:", err)
			break
		}
		log.Printf("recv: %s", message)

		if string(message) == "Test packet" {
			wsMsg := &gw.WSMessage{Method: "test", Path: "/test", Body: message}
			answer, err := protojson.Marshal(wsMsg)
			if err != nil {
				log.Println("marshalTestAnswer:", err)
				break
			}
			err = c.WriteMessage(mt, answer)
			if err != nil {
				log.Println("write:", err)
				break
			}
			continue
		}

		wsMsg := &gw.WSMessage{}
		//err = proto.Unmarshal(message, wsMsg)
		err = protojson.Unmarshal(message, wsMsg)
		if err != nil {
			log.Println("unmarshal WSMessage:", err)
			break
		}

		log.Println("crafting request:", wsMsg)
		// forward request to REST Gateway
		request, err := http.NewRequest(wsMsg.Method, wsMsg.Path, bytes.NewReader(wsMsg.Body))
		if err != nil {
			log.Println("crafting request:", err)
			break
		}

		log.Println("forwarding request")
		rw := NewResponseWriter()
		wsg.mux.ServeHTTP(rw, request)

		log.Println("sending response")
		wsMsg.Body = rw.GetData()
		answer, err := protojson.Marshal(wsMsg)
		if err != nil {
			log.Println("answer marshal:", err)
			break
		}
		// send response back to ws client
		err = c.WriteMessage(mt, answer)
		if err != nil {
			log.Println("write:", err)
			break
		}
	}
}

func RegisterWebsocketHandler(ctx context.Context, mux *runtime.ServeMux, endpoint string, opts []grpc.DialOption) {
	wsGateway := WSGateway{mux: mux}
	mux.HandlePath("GET", "/ws", wsGateway.Handler)
	mux.HandlePath("GET", "/ws/demo", demo)
	log.Println("added WSGateway handler")
}

func demo(w http.ResponseWriter, r *http.Request, pathParams map[string]string) {
	demoTemplate.Execute(w, "ws://"+r.Host+"/ws")
}

var demoTemplate = template.Must(template.New("").Parse(`
<!DOCTYPE html>
<html>
<head>
<meta charset="utf-8">
<script>  
window.addEventListener("load", function(evt) {
    var output = document.getElementById("output");
    var input = document.getElementById("input");
    var ws;
    var print = function(message) {
        var d = document.createElement("div");
        d.textContent = message;
        output.appendChild(d);
        output.scroll(0, output.scrollHeight);
    };
    document.getElementById("open").onclick = function(evt) {
        if (ws) {
            return false;
        }
        ws = new WebSocket("{{.}}");
        ws.onopen = function(evt) {
            print("OPEN");
        }
        ws.onclose = function(evt) {
            print("CLOSE");
            ws = null;
        }
        ws.onmessage = function(evt) {
            print("RESPONSE: " + evt.data);
        }
        ws.onerror = function(evt) {
            print("ERROR: " + evt.data);
        }
        return false;
    };
    document.getElementById("send").onclick = function(evt) {
        if (!ws) {
            return false;
        }
        print("SEND: " + input.value);
        ws.send(input.value);
        return false;
    };
    document.getElementById("close").onclick = function(evt) {
        if (!ws) {
            return false;
        }
        ws.close();
        return false;
    };
});
</script>
</head>
<body>
<table>
<tr><td valign="top" width="50%">
<p>Click "Open" to create a connection to the server, 
"Send" to send a message to the server and "Close" to close the connection. 
You can change the message and send multiple times.
<p>
<form>
<button id="open">Open</button>
<button id="close">Close</button>
<p><input id="input" type="text" value="Hello world!">
<button id="send">Send</button>
</form>
</td><td valign="top" width="50%">
<div id="output" style="max-height: 70vh;overflow-y: scroll;"></div>
</td></tr></table>
</body>
</html>
`))
